/* -*-c-*-
 *
 * $Id: rsa-gen.c,v 1.1 1999/12/22 15:50:45 mdw Exp $
 *
 * RSA parameter generation
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: rsa-gen.c,v $
 * Revision 1.1  1999/12/22 15:50:45  mdw
 * Initial RSA support.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/dstr.h>

#include "grand.h"
#include "mp.h"
#include "pgen.h"
#include "rsa.h"
#include "strongprime.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @rsa_gen@ --- *
 *
 * Arguments:	@rsa_param *rp@ = pointer to block to be filled in
 *		@unsigned nbits@ = required modulus size in bits
 *		@grand *r@ = random number source
 *		@unsigned n@ = number of attempts to make
 *		@pgen_proc *event@ = event handler function
 *		@void *ectx@ = argument for the event handler
 *
 * Returns:	Zero if all went well, nonzero otherwise.
 *
 * Use:		Constructs a pair of strong RSA primes and other useful RSA
 *		parameters.  A small encryption exponent is chosen if
 *		possible.
 */

int rsa_gen(rsa_param *rp, unsigned nbits, grand *r, unsigned n,
	    pgen_proc *event, void *ectx)
{
  mp *phi;
  int i;
  mp *g;

  /* --- Generate strong primes %$p$% and %$q$% --- */

  if ((rp->p = strongprime("p", MP_NEW, nbits/2, r, n, event, ectx)) == 0)
    goto fail_p;
  if ((rp->q = strongprime("q", MP_NEW, nbits/2, r, n, event, ectx)) == 0)
    goto fail_q;

  /* --- Work out the modulus and the CRT coefficient --- */

  rp->n = mp_mul(MP_NEW, rp->p, rp->q);
  rp->q_inv = MP_NEW; mp_gcd(0, 0, &rp->q_inv, rp->p, rp->q);

  /* --- Work out %$\varphi(n) = (p - 1)(q - 1)$% --- *
   *
   * Save on further multiplications by noting that %$n = pq$% is known and
   * that %$(p - 1)(q - 1) = pq - p - q + 1$%.
   */

  phi = mp_sub(MP_NEW, rp->n, rp->p);
  phi = mp_sub(phi, phi, rp->q);
  phi = mp_add(phi, phi, MP_ONE);

  /* --- Decide on a public exponent --- *
   *
   * Simultaneously compute the private exponent.
   */

  rp->e = mp_create(1);
  rp->d = MP_NEW;
  g = MP_NEW;
  for (i = 1; i < NPRIME; i++) {
    rp->e->v[0] = primetab[i];
    mp_gcd(&g, 0, &rp->d, phi, rp->e);
    if (MP_CMP(g, ==, MP_ONE))
      goto good_e;
  }
  goto fail_e;

  /* --- Work out exponent residues --- */

good_e:
  rp->dp = MP_NEW; phi = mp_sub(phi, rp->p, MP_ONE);
  mp_div(0, &rp->dp, rp->d, phi);

  rp->dq = MP_NEW; phi = mp_sub(phi, rp->q, MP_ONE);
  mp_div(0, &rp->dq, rp->d, phi);

  /* --- Done --- */

  mp_drop(phi);
  mp_drop(g);
  return (0);

  /* --- Tidy up when something goes wrong --- */

fail_e:
  mp_drop(g);
  mp_drop(phi);
  mp_drop(rp->n);
  mp_drop(rp->q_inv);
  mp_drop(rp->q);
fail_q:
  mp_drop(rp->p);
fail_p:
  return (-1);
}

/*----- That's all, folks -------------------------------------------------*/
